#ifndef _FONTFILEST_H_
#define _FONTFILEST_H_

/**
 * fontxlfd.h
 **/
typedef struct _FontScalable
{
	int pixel, point, x, y, width;
	double pixel_matrix[4], point_matrix[4];
	int pixel_set, point_set, horiz_weight, vert_weight;
	double ShearSin, ShearCos, RotateSin, RotateCos;
	Bool horiz_mirror, vert_mirror;
}
FontScalableRec, *FontScalablePtr;

#define FONT_XLFD_REPLACE_NONE	0
#define FONT_XLFD_REPLACE_STAR	1
#define FONT_XLFD_REPLACE_ZERO	2
#define FONT_XLFD_REPLACE_VALUE	3


/**
 * fontfile.h
 **/
typedef struct _FontEntry *FontEntryPtr;
typedef struct _FontTable *FontTablePtr;
typedef struct _FontName *FontNamePtr;
typedef struct _FontScaled *FontScaledPtr;
typedef struct _FontScalableExtra *FontScalableExtraPtr;
typedef struct _FontScalableEntry *FontScalableEntryPtr;
typedef struct _FontScaleAliasEntry *FontScaleAliasEntryPtr;
typedef struct _FontBitmapEntry *FontBitmapEntryPtr;
typedef struct _FontAliasEntry *FontAliasEntryPtr;
typedef struct _FontBCEntry *FontBCEntryPtr;
typedef struct _FontDirectory *FontDirectoryPtr;
typedef struct _FontRenderer *FontRendererPtr;

#define NullFontEntry		    ((FontEntryPtr) 0)
#define NullFontTable		    ((FontTablePtr) 0)
#define NullFontName		    ((FontNamePtr) 0)
#define NullFontScaled		    ((FontScaled) 0)
#define NullFontScalableExtra	    ((FontScalableExtra) 0)
#define NullFontscalableEntry	    ((FontScalableEntry) 0)
#define NullFontScaleAliasEntry	    ((FontScaleAliasEntry) 0)
#define NullFontBitmapEntry	    ((FontBitmapEntry) 0)
#define NullFontAliasEntry	    ((FontAliasEntry) 0)
#define NullFontBCEntry		    ((FontBCEntry) 0)
#define NullFontDirectory	    ((FontDirectoryPtr) 0)
#define NullFontRenderer	    ((FontRendererPtr) 0)

#define FONT_ENTRY_SCALABLE	0
#define FONT_ENTRY_SCALE_ALIAS	1
#define FONT_ENTRY_BITMAP	2
#define FONT_ENTRY_ALIAS	3
#define FONT_ENTRY_BC		4

#define MAXFONTNAMELEN	    1024
#define MAXFONTFILENAMELEN  1024

#define FontDirFile	        "fonts.dir"
#define FontAliasFile       "fonts.alias"
#define FontScalableFile    "fonts.scale"

typedef struct _FontName
{
	char *name;
	short length;
	short ndashes;
}
FontNameRec;

typedef struct _FontScaled
{
	FontScalableRec vals;
	FontEntryPtr bitmap;
	FontPtr pFont;
} FontScaledRec;

typedef struct _FontScalableExtra
{
	FontScalableRec defaults;
	int numScaled;
	int sizeScaled;
	FontScaledPtr scaled;
	pointer private;
}
FontScalableExtraRec;

typedef struct _FontScalableEntry
{
	FontRendererPtr renderer;
	char *fileName;
	FontScalableExtraPtr extra;
}
FontScalableEntryRec;


/*
 * This "can't" work yet - the returned alias string must be permanent,
 * but this layer would need to generate the appropriate name from the
 * resolved scalable + the XLFD values passed in.  XXX
 */
typedef struct _FontScaleAliasEntry
{
	char *resolved;
}
FontScaleAliasEntryRec;

typedef struct _FontBitmapEntry
{
	FontRendererPtr renderer;
	char *fileName;
	FontPtr pFont;
}
FontBitmapEntryRec;

typedef struct _FontAliasEntry
{
	char *resolved;
}
FontAliasEntryRec;

typedef struct _FontBCEntry
{
	FontScalableRec vals;
	FontEntryPtr entry;
} FontBCEntryRec;

typedef struct _FontEntry
{
	FontNameRec name;
	int type;
	union _FontEntryParts
	{
		FontScalableEntryRec scalable;
		FontBitmapEntryRec bitmap;
		FontAliasEntryRec alias;
		FontBCEntryRec bc;
	} u;
} FontEntryRec;

typedef struct _FontTable
{
	int used;
	int size;
	FontEntryPtr entries;
	Bool sorted;
}
FontTableRec;

typedef struct _FontDirectory
{
	char *directory;
	unsigned long dir_mtime;
	unsigned long alias_mtime;
	FontTableRec scalable;
	FontTableRec nonScalable;
}
FontDirectoryRec;

typedef struct _FontRenderer
{
	char *fileSuffix;
	int fileSuffixLen;
	int (*OpenBitmap) ( /* fpe, pFont, flags, entry, fileName, format, fmask */ );
	int (*OpenScalable) ( /* fpe, pFont, flags, entry, fileName, vals, format, fmask */ );
	int (*GetInfoBitmap) ( /* fpe, pFontInfo, entry, fileName */ );
	int (*GetInfoScalable) ( /* fpe, pFontInfo, entry, fileName, vals */ );
	int number;
}
FontRendererRec;

typedef struct _FontRenders
{
	int number;
	FontRendererPtr *renderers;
}
FontRenderersRec, *FontRenderersPtr;

typedef struct _BitmapInstance
{
	FontScalableRec vals;
	FontBitmapEntryPtr bitmap;
} BitmapInstanceRec, *BitmapInstancePtr;

typedef struct _BitmapScalablePrivate
{
	int numInstances;
	BitmapInstancePtr instances;
}
BitmapScalablePrivateRec, *BitmapScalablePrivatePtr;

typedef struct _BitmapSources
{
	FontPathElementPtr *fpe;
	int size;
	int count;
}
BitmapSourcesRec, *BitmapSourcesPtr;

#endif /* _FONTFILEST_H_ */
